import os
from safetensors import safe_open
from safetensors.torch import save_file
from exllamav2.model import ExLlamaV2Attention, ExLlamaV2MLP, ExLlamaV2RMSNorm, ExLlamaV2Embedding, ExLlamaV2Linear


def get_f_module(job, module):

    mod_dict = {}
    module.load()
    mod_dict[module.key + ".weight"] = module.get_weight()
    return mod_dict


def get_q_module(job, module):

    mod_dict = {}
    filename = os.path.join(job["out_dir"], "out_tensor/" + module.key + ".safetensors")
    with safe_open(filename, framework = "pt", device = "cpu") as f:
        for k in f.keys():
            mod_dict[k] = f.get_tensor(k)
    return mod_dict


def compile_model(job, save_fn, model):

    out_dict = {}

    index = 0
    while index < len(model.modules):

        module = model.modules[index]

        if isinstance(module, ExLlamaV2Embedding):

            out_dict |= get_f_module(job, module)

        if isinstance(module, ExLlamaV2Attention):

            out_dict |= get_f_module(job, module.input_layernorm)
            out_dict |= get_q_module(job, module.q_proj)
            out_dict |= get_q_module(job, module.k_proj)
            out_dict |= get_q_module(job, module.v_proj)
            out_dict |= get_q_module(job, module.o_proj)

        if isinstance(module, ExLlamaV2MLP):

            out_dict |= get_f_module(job, module.post_attention_layernorm)
            out_dict |= get_q_module(job, module.gate_proj)
            out_dict |= get_q_module(job, module.up_proj)
            out_dict |= get_q_module(job, module.down_proj)

        if isinstance(module, ExLlamaV2RMSNorm):

            out_dict |= get_f_module(job, module)

        if isinstance(module, ExLlamaV2Linear):

            assert module.key == "lm_head"
            out_dict |= get_q_module(job, module)

        index += 1

    out_file = os.path.join(job["out_dir"], "output.safetensors")
    save_file(out_dict, out_file)



def optimize(job, save_fn):

    numel = 0
    max_rfn = 0.0
    for layer in job["measurement"]:
        numel += layer["numel"]
        for option in layer["options"]:
            max_rfn = max(max_rfn, option["err"])

    # Find minimum error that satisfies target size

    target_bpw = job["bits"]
    target_rfn = max_rfn

    target_rfn_step = max_rfn / 2

    while True:

        current_total_bits = 0
        for layer in job["measurement"]:

            best_option = None
            best_bpw = 100.0

            for option in layer["options"]:
                if option["bpw"] < best_bpw and option["err"] <= target_rfn:
                    best_bpw = option["bpw"]
                    best_option = option

            if best_option is not None:
                layer["best_option"] = best_option

            current_total_bits += int(layer["best_option"]["total_bits"])

        current_bpw = current_total_bits / numel
        print(f" -- rfn_error: {target_rfn:2.5f}  bpw: {current_bpw:2.5f}")

        eps = 1e-3

        if current_bpw < target_bpw - eps:
            target_rfn -= target_rfn_step
        elif  current_bpw > target_bpw + eps:
            target_rfn += target_rfn_step

        target_rfn_step /= 2
        if target_rfn_step < 1e-6: break

